from __future__ import annotations

import io
import os
from pathlib import Path

from lxml import etree

from pygef.broxml import resolvers
from pygef.broxml.xml_parser import BaseParser, read_xml
from pygef.cpt import CPTData

# maps keyword argument to:
# xpath: query passed to elementree.find
# resolver: A function that converts the string to the proper datatype
#           Fn(str) -> Any
# el-atrr: Optional: attribute of an element taken before send to resolver
CPT_ATTRIBS = {
    "bro_id": {"xpath": "brocom:broId"},
    "research_report_date": {
        "xpath": "./researchReportDate/brocom:date",
        "resolver": resolvers.parse_date,
        "el-attr": "text",
    },
    "cpt_standard": {"xpath": "./cptStandard"},
    "delivered_location": {
        "xpath": "./deliveredLocation/cptcommon:location",
        "resolver": resolvers.parse_gml_location,
    },
    "standardized_location": {
        "xpath": "./standardizedLocation/brocom:location",
        "resolver": resolvers.parse_gml_location,
    },
    "dissipationtest_performed": {
        "xpath": "./conePenetrometerSurvey/cptcommon:dissipationTestPerformed",
        "resolver": resolvers.parse_bool,
        "el-attr": "text",
    },
    "quality_class": {
        "xpath": "./conePenetrometerSurvey/cptcommon:qualityClass",
        "resolver": resolvers.parse_quality_class,
        "el-attr": "text",
    },
    "groundwater_level": {
        "xpath": "./additionalInvestigation/cptcommon:groundwaterLevel",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "predrilled_depth": {
        "xpath": "./conePenetrometerSurvey/cptcommon:trajectory/cptcommon:predrilledDepth",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "final_depth": {
        "xpath": "./conePenetrometerSurvey/cptcommon:trajectory/cptcommon:finalDepth",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "data": {
        "xpath": "./conePenetrometerSurvey",
        "resolver": resolvers.process_cpt_result,
    },
    "cpt_description": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:description",
    },
    "cpt_type": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:conePenetrometerType",
    },
    "cone_surface_area": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:coneSurfaceArea",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "cone_diameter": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:coneDiameter",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "cone_surface_quotient": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:coneSurfaceQuotient",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "cone_to_friction_sleeve_distance": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:coneToFrictionSleeveDistance",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "cone_to_friction_sleeve_surface_area": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:frictionSleeveSurfaceArea",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "cone_to_friction_sleeve_surface_quotient": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:frictionSleeveSurfaceQuotient",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_cone_resistance_before": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:coneResistanceBefore",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_cone_resistance_after": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:coneResistanceAfter",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_inclination_ew_before": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:inclinationEWBefore",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "zlm_inclination_ew_after": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:inclinationEWAfter",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "zlm_inclination_ns_before": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:inclinationNSBefore",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "zlm_inclination_ns_after": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:inclinationNSAfter",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "zlm_inclination_resultant_before": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:inclinationResultantBefore",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "zlm_inclination_resultant_after": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:inclinationResultantAfter",
        "resolver": resolvers.parse_int,
        "el-attr": "text",
    },
    "zlm_local_friction_before": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:localFrictionBefore",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_local_friction_after": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:localFrictionAfter",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_pore_pressure_u1_before": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:porePressureU1Before",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_pore_pressure_u2_before": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:porePressureU2Before",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_pore_pressure_u3_before": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:porePressureU3Before",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_pore_pressure_u1_after": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:porePressureU1After",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_pore_pressure_u2_after": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:porePressureU2After",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "zlm_pore_pressure_u3_after": {
        "xpath": "./conePenetrometerSurvey/cptcommon:conePenetrometer/cptcommon:zeroLoadMeasurement/cptcommon:porePressureU3After",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "delivered_vertical_position_offset": {
        "xpath": "./deliveredVerticalPosition/cptcommon:offset",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "delivered_vertical_position_datum": {
        "xpath": "./deliveredVerticalPosition/cptcommon:verticalDatum",
        "resolver": resolvers.parse_local_vertical_reference_point_class,
        "el-attr": "text",
    },
    "delivered_vertical_position_reference_point": {
        "xpath": "./deliveredVerticalPosition/cptcommon:localVerticalReferencePoint",
        "resolver": resolvers.lower_text,
        "el-attr": "text",
    },
}


def read_cpt(file: io.BytesIO | Path | str) -> list[CPTData]:
    if isinstance(file, str) and not os.path.exists(file):
        root = etree.fromstring(file, parser=BaseParser).getroot()
    else:
        root = etree.parse(file, parser=BaseParser).getroot()
    return read_xml(root, CPTData, CPT_ATTRIBS, "dispatchDocument")



from __future__ import annotations


def concat_args(msg: str, args: tuple):
    if not isinstance(args, tuple):
        args = tuple(args)
    return " ".join(str(x) for x in [msg, *args])


class UserError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class ParseGefError(UserError):
    def __init__(self, *args, **kwargs):
        msg = "Pygef encountered an error during parsing of the .gef file"
        UserError.__init__(self, concat_args(msg, args), **kwargs)


class ParseCptGefError(ParseGefError):
    def __init__(self, *args, **kwargs):
        msg = "for CPT data"
        UserError.__init__(self, concat_args(msg, args), **kwargs)



from __future__ import annotations

import polars as pl

from pygef.gef import utils
from pygef.gef.gef import _Gef, parse_all_columns_info, replace_column_void
from pygef.gef.mapping import MAP_QUANTITY_NUMBER_COLUMN_NAME_BORE


class _GefBore(_Gef):
    def __init__(self, path=None, string=None):
        """
        Parser of the borehole file.

        Parameters
        ----------
        path: str
            Path to the *.gef file.
        string: str
            String version of the *.gef file.
        """
        super().__init__(path=path, string=string)
        if self.type == "bore":
            pass
        elif self.type == "borehole-report":
            raise ValueError(
                "The selected gef file is a GEF-BOREHOLE-Report. Can only parse "
                "GEF-CPT-Report and GEF-BORE-Report. Check the PROCEDURECODE."
            )
        else:
            raise ValueError(
                "The selected gef file is not a borehole. "
                "Check the REPORTCODE or the PROCEDURECODE."
            )

        self.project_id = utils.parse_project_type(self._headers, "bore")
        self.nen_version = "NEN 5104"

        self.data_info = self.ColumnsInfo(
            *parse_all_columns_info(
                self._headers, MAP_QUANTITY_NUMBER_COLUMN_NAME_BORE
            ),
            col_separator=utils.get_column_separator(self._headers),
            rec_separator=utils.get_record_separator(self._headers),
            column_voids=utils.parse_column_void(self._headers),
        )

        data_s_rows = self._data.split(self.data_info.rec_separator)
        data_rows_soil = self.extract_soil_info(
            data_s_rows, self.data_info.columns_number, self.data_info.col_separator
        )

        self.df = (
            self.parse_data(
                self._data,
                self.data_info.col_separator,
                self.data_info.rec_separator,
                self.data_info.descriptions,
            )
            .pipe(replace_column_void, self.data_info.description_to_void_mapping)
            .pipe(self.parse_data_soil_code, data_rows_soil)
            .pipe(self.parse_add_info_as_string, data_rows_soil)
            .pipe(self.map_soil_code_to_soil_name, data_rows_soil)
        )

        # Remove the rows with null values
        self.df.drop_nulls()

    @staticmethod
    def parse_add_info_as_string(
        df: pl.DataFrame, data_rows_soil: list[list[str]]
    ) -> pl.DataFrame:
        return df.with_columns(
            pl.Series(
                "remarks",
                [utils.parse_add_info("".join(row[1::])) for row in data_rows_soil],
            )
        )

    @staticmethod
    def extract_soil_info(data_s_rows, columns_number, column_separator):
        return list(
            map(
                lambda x: x.split(column_separator)[columns_number:-1], data_s_rows[:-1]
            )
        )

    @staticmethod
    def parse_data_soil_code(df: pl.DataFrame, data_rows_soil: list[list[str]]):
        # return df.with_columns(pl.Series("soil_code", data_rows_soil).apply(utils.parse_soil_code))
        return df.with_columns(
            pl.Series(
                "geotechnicalSoilCode",
                list(map(lambda x: utils.parse_soil_code(x[0]), data_rows_soil)),
            )
        )

    @staticmethod
    def map_soil_code_to_soil_name(df: pl.DataFrame, data_rows_soil: list[str]):
        # return df.with_columns(pl.Series("soil_code", data_rows_soil).apply(utils.parse_soil_code))
        return df.with_columns(
            pl.Series(
                "geotechnicalSoilName",
                list(
                    map(
                        lambda x: utils.parse_soil_name(utils.parse_soil_code(x[0])),
                        data_rows_soil,
                    )
                ),
            )
        )



from __future__ import annotations

import logging
import re
from datetime import date
from typing import Dict, List, Tuple, Union

import numpy as np

from pygef import exceptions
from pygef.gef.mapping import MAPPING_PARAMETERS

logger = logging.getLogger(__name__)


def cast_string(f, s):
    """
    Generic function that casts a string.

    :param f: (function) Any casting function.
    :param s: (str) String to cast.
    :return: The casted object
    """
    try:
        return f(s)
    except ValueError:
        return None


def measurement_header_value(headers, name, index=0, cast=None, key=0):
    """
    Based on the key en index of the measurement in
    the GEF return the value

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param name:(str) Header name.
    :param key:(int) Index of the variable.
    :param index:(int) Index of the value.
    :param cast: (function) Cast to the specified type.
    :return: (str) The header value.
    """
    if name in headers:
        mapping = {int(value[0]): value for value in headers[name]}

        if key in mapping:
            result = mapping[key][index]
            if cast:
                return cast(result)
            else:
                return result


def first_header_value(headers, name, index=0, cast=None):
    """
    Get the first matching line with a header.

    Throws an error when the header exists but doesn't have the value at the
    passed index.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param name:(str) Header name.
    :param index:(int) Index of the value.
    :param cast: (function) Cast to the specified type.
    :return: (str) The header value.
    """
    if name in headers:
        result = headers[name][0][index]
        if cast:
            return cast(result)
        else:
            return result


def parse_regex_cast(regex_string, s, f, group_number):
    """
    Function that searches a regex match and casts the result.

    :param regex_string: (str) Regex pattern.
    :param s: (str) String to search for regex pattern.
    :param f: (function) To apply to the regex match.
    :param group_number: (int) Which group number to query when there is a match.
    :return: Casted result.
    """
    g = re.search(regex_string, s)
    if g:
        return cast_string(f, g.group(group_number))
    else:
        return None


def parse_column_void(headers: Union[dict, str]) -> Dict[int, float]:
    """
    Function that parses the column void headers and returns a dictionary that maps
    column-numbers to their void value.

    Returns a ParseGefError on duplicates and ill header formats
    Does not guarantee that all columns have a void value.

    :param headers:(Union[Dict,str]) Gef headers
    :return: (Dict[int, float]) Mapping of column-numbers to their void value.
    """
    voids_info: List[Tuple[int, float]] = []
    if isinstance(headers, dict):
        # Return a list of all the second float values of all COLUMN_VOID lines
        if "COLUMNVOID" in headers:
            try:
                voids_info = list(
                    map(
                        lambda values: (int(values[0]), float(values[1])),
                        headers["COLUMNVOID"],
                    )
                )
            except ValueError:
                raise exceptions.ParseGefError(
                    ": One of more #COLUMNVOID headers have an invalid format."
                )

    else:
        for void_line in re.finditer(r"#COLUMNVOID\s*=\s*(.*)", headers):
            voids = re.search(r"^(\d+)\s*,\s*([-+]?\d+\.?\d*)", void_line.group(1))

            if not voids:
                raise exceptions.ParseGefError(
                    ": One of more #COLUMNVOID headers have an invalid format."
                )

            voids_info.append((int(voids.group(1)), float(voids.group(2))))

    col_numbers = list(map(lambda values: values[0], voids_info))
    if any(np.unique(col_numbers, return_counts=True)[1] > 1):
        raise exceptions.ParseGefError(
            ": One or more #COLUMNVOID headers have duplicate definitions."
        )

    column_void = {item[0]: item[1] for item in voids_info}

    return column_void


def parse_measurement_var_as_float(headers, var_number):
    """
    Function that returns a measurement variable as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param var_number: (int) Variable number.
    :return: Variable value.
    """
    # Convert the variable number to a string so it's cheaper to compare
    var_number_str = str(var_number)

    try:
        if isinstance(headers, dict):
            # Loop over all headers to find the right number
            if "MEASUREMENTVAR" in headers:
                for values in headers["MEASUREMENTVAR"]:
                    if values[0] == var_number_str:
                        return float(values[1])
        else:
            # Find all '#MEASUREMENTVAR= **,' strings first
            for match in re.finditer(
                r"#MEASUREMENTVAR[=\s+]+(\d+)[, ]+([\d-]+\.?\d*)", headers
            ):
                # The first group is the variable number
                if match.group(1) == var_number_str:
                    # The second group is the actual value
                    return cast_string(float, match.group(2))

    except ValueError:
        pass


def parse_cone_id(headers):
    """
    Function that returns the cone id specified in GEF-CPT file.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: cone id.
    """
    if isinstance(headers, dict):
        return measurement_header_value(headers, "MEASUREMENTTEXT", index=1, key=4)
    else:
        try:
            return parse_regex_cast(
                r"#MEASUREMENTTEXT[=\s+]+4[, ]+([\w.-]+)", headers, str, 1
            )
        except ValueError:
            return None


def parse_cpt_class(headers) -> int | None:
    """
    Function that returns the class of the cpt as an int.
    The word class or klasse has to be defined in the parsed string.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: Cpt class.
    """
    if isinstance(headers, dict):
        all_definition = measurement_header_value(
            headers, "MEASUREMENTTEXT", key=6, index=1
        )
    else:
        all_definition = parse_regex_cast(
            r"#MEASUREMENTTEXT[=\s+]+6[, ](.*)", headers, str, 1
        )
    if all_definition is not None:
        return parse_regex_cast(
            r"^.*?(klasse|class|kl.).*?(\d{1})", all_definition.lower(), int, 2
        )
    return None


def parse_project_type(headers, gef_type):
    """
    Function that returns the project type as an int.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param gef_type: (str) String from which the gef type is given.
    :return: Project type number.
    """
    if isinstance(headers, dict):
        if gef_type == "cpt":
            try:
                # Try to get the second value first
                return first_header_value(headers, "PROJECTID", index=1)
            except Exception:
                # If that fails get the first one
                return first_header_value(headers, "PROJECTID")
        elif gef_type == "bore":
            return first_header_value(headers, "PROJECTID")
    else:
        if gef_type == "cpt":
            return parse_regex_cast(r"PROJECTID[\s=a-zA-Z,]*([\w-]+)", headers, str, 1)
        elif gef_type == "bore":
            return parse_regex_cast(r"#PROJECTID+[^a-zA-Z]+([\w-]+)", headers, str, 1)


def parse_zid_as_float(headers):
    """
    Function that returns the zid as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:(float) ZID number.
    """
    if isinstance(headers, dict):
        if "ZID" in headers:
            return first_header_value(headers, "ZID", index=1, cast=float)
    else:
        return parse_regex_cast(
            r"#ZID[=\s+]+[^,]*[,\s+]+([^?!,$|\s$]+)", headers, float, 1
        )


def parse_height_system(headers):
    """
    Function that returns the zid as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:(float) ZID number.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "ZID", cast=float)
    else:
        return parse_regex_cast(r"#ZID[=\s+]+([^,]*)", headers, float, 1)


def parse_xid_as_float(headers):
    """
    Function that returns the x coordinate as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:(float) x coordinate value.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "XYID", index=1, cast=float)
    else:
        return parse_regex_cast(
            r"#XYID[=\s+]*.*?,\s*(\d*\s*(\.|\d)*),\s*(\d*(\.|\d)*)", headers, float, 1
        )


def parse_yid_as_float(headers):
    """
    Function that returns the y coordinate as a float.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:(float) y coordinate value.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "XYID", index=2, cast=float)
    else:
        return parse_regex_cast(
            r"#XYID[=\s+]*.*?,\s*(\d*\s*(\.|\d)*),\s*(\d*(\.|\d)*)", headers, float, 3
        )


def parse_gef_type(headers):
    """
    Function that returns the gef type.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: (str) Gef type.
    """
    proc_code = ""
    if isinstance(headers, dict):
        if "REPORTCODE" in headers:
            proc_code = first_header_value(headers, "REPORTCODE").lower()
        elif "PROCEDURECODE" in headers:
            proc_code = first_header_value(headers, "PROCEDURECODE").lower()
        else:
            return None
    else:
        if parse_regex_cast(
            r"#PROCEDURECODE[^a-zA-Z]+([\w-]+)", headers, lambda x: x.lower(), 1
        ) and parse_regex_cast(
            r"#REPORTCODE[^a-zA-Z]+([\w-]+)", headers, lambda x: x.lower(), 1
        ):
            proc_code = parse_regex_cast(
                r"#REPORTCODE[^a-zA-Z]+([\w-]+)", headers, lambda x: x.lower(), 1
            )
        else:
            proc_code = parse_regex_cast(
                r"#(REPORTCODE|PROCEDURECODE)[^a-zA-Z]+([\w-]+)",
                headers,
                lambda x: x.lower(),
                2,
            )

    if "cpt" in proc_code or "dis" in proc_code:
        return "cpt"
    elif "bore" in proc_code and "borehole" not in proc_code:
        return "bore"
    elif "borehole" in proc_code:
        return "borehole-report"


def parse_file_date(headers):
    """
    Fuction to parse the file date.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: File date.
    """
    year = None
    month = None
    day = None

    if isinstance(headers, dict):
        if "FILEDATE" in headers:
            year = first_header_value(headers, "FILEDATE", index=0, cast=int)
            month = first_header_value(headers, "FILEDATE", index=1, cast=int)
            day = first_header_value(headers, "FILEDATE", index=2, cast=int)
        else:
            return None
    else:
        g = re.search(r"#FILEDATE[\s=]*(\d+)[,\s+]+(\d+)[,\s+]+(\d+)", headers)
        if g:
            year = int(g.group(1))
            month = int(g.group(2))
            day = int(g.group(3))
        else:
            return None

    return date(year, month, day)


def parse_columns_number(headers: Union[dict, str]) -> int:
    """
    Function that returns the columns number as an int.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: Columns number value.
    """
    if isinstance(headers, dict):
        if "COLUMNINFO" in headers:
            return len(headers["COLUMNINFO"])

    else:
        col_numbers = re.findall(r"#COLUMNINFO[=\s]+(\d+)", headers)
        if col_numbers:
            return len(col_numbers)

    return 0


def get_description(
    quantity_number: int,
    custom_description: str,
    dictionary: Dict[int, str] | None = None,
) -> str:
    """
    Returns the default description of a quantity number if available in the
    provided dictionary, else returns the custom_description.

    :param quantity_number: (int) The quantity number of the column
    :param custom_description: (str) The column description from the .gef file
    :param dictionary: (Optional[Dict[int, str]]) The dictionary that maps
        quantity-numbers to descriptions
    :return: (str) The column description
    """

    if dictionary is not None and quantity_number in dictionary:
        return dictionary[quantity_number]
    return custom_description


def parse_column_separator(headers):
    """
    Function to parse the column separator. It is used only in the borehole class.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: Column separator.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "COLUMNSEPARATOR")
    else:
        return parse_regex_cast(r"#COLUMNSEPARATOR+[=\s+]+(.)", headers, str, 1)


def parse_test_id(headers):
    """
    Function to parse the test id.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: test id.
    """
    result = None
    if isinstance(headers, dict):
        result = first_header_value(headers, "TESTID")
    else:
        result = parse_regex_cast(r"#TESTID+[=\s+]+(.*)", headers, str, 1)

    if result is not None:
        return result.strip()


def parse_record_separator(headers):
    """
    Function to parse the record separator(end of the line). It is used only in the borehole class.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: Record separator.
    """
    if isinstance(headers, dict):
        return first_header_value(headers, "RECORDSEPARATOR")
    else:
        return parse_regex_cast(r"#RECORDSEPARATOR+[=\s+]+(.)", headers, str, 1)


def get_column_separator(headers: Union[dict, str]) -> str:
    """
    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:
    """
    return parse_column_separator(headers) or " "


def get_record_separator(headers: Union[dict, str]) -> str:
    """
    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return:
    """
    return parse_record_separator(headers) or "\n"


def parse_soil_code(s: str) -> str:
    """
    Function to parse the soil code.

    :param s: (str) String with the soil code.
    :return: Soil code.
    """
    return s.replace("'", "")


def parse_soil_name(s: str) -> str:
    """
    Function to parse the soil name.

    NOTE: interpretation of the soil code to NEN-EN-ISO 14688-1:2019+NEN 8990:2020

    :param s: (str) String with the soil code.
    :return: Soil name.
    """
    return MAPPING_PARAMETERS.dino_to_bro(s)


def parse_add_info(headers):
    """
    Function to parse all the additional informations.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :return: (str) Additional informations.
    """
    dict_add_info = MAPPING_PARAMETERS.code_to_text()

    if isinstance(headers, dict):
        raise Exception("todo")
    else:
        string_noquote = headers[1:-1]
        string2list = string_noquote.split("''")
        add_info = ""
        for i, string in enumerate(string2list):
            if string:
                add_info += "{}) ".format(i + 1) + "".join(
                    [
                        (
                            dict_add_info[string_split]
                            if string_split in dict_add_info
                            else string_split + " "
                        )
                        for string_split in string.split(" ")
                    ]
                )
        return add_info



from __future__ import annotations

from typing import List

import numpy as np
import polars as pl

from pygef.gef import utils
from pygef.gef.gef import _Gef, parse_all_columns_info, replace_column_void
from pygef.gef.mapping import MAP_QUANTITY_NUMBER_COLUMN_NAME_CPT


class _GefCpt(_Gef):
    def __init__(self, path=None, string=None):
        """
        Parser of the cpt file.

        Parameters
        ----------
        path: str
            Path to the *.gef file.
        string: str
            String version of the *.gef file.
        """
        super().__init__(path=path, string=string)
        if not self.type == "cpt":
            raise ValueError(
                "The selected gef file is not a cpt. "
                "Check the REPORTCODE or the PROCEDURECODE."
            )
        self.project_id = utils.parse_project_type(self._headers, "cpt")
        self.cone_id = utils.parse_cone_id(self._headers)

        self.cpt_class = utils.parse_cpt_class(self._headers)

        self.nom_surface_area_cone_tip = utils.parse_measurement_var_as_float(
            self._headers, 1
        )
        self.nom_surface_area_friction_element = utils.parse_measurement_var_as_float(
            self._headers, 2
        )
        self.net_surface_area_quotient_of_the_cone_tip = (
            utils.parse_measurement_var_as_float(self._headers, 3)
        )
        self.net_surface_area_quotient_of_the_friction_casing = (
            utils.parse_measurement_var_as_float(self._headers, 4)
        )
        self.distance_between_cone_and_centre_of_friction_casing = (
            utils.parse_measurement_var_as_float(self._headers, 5)
        )
        self.friction_present = utils.parse_measurement_var_as_float(self._headers, 6)
        self.ppt_u1_present = utils.parse_measurement_var_as_float(self._headers, 7)
        self.ppt_u2_present = utils.parse_measurement_var_as_float(self._headers, 8)
        self.ppt_u3_present = utils.parse_measurement_var_as_float(self._headers, 9)
        self.inclination_measurement_present = utils.parse_measurement_var_as_float(
            self._headers, 10
        )
        self.use_of_back_flow_compensator = utils.parse_measurement_var_as_float(
            self._headers, 11
        )
        self.type_of_cone_penetration_test = utils.parse_measurement_var_as_float(
            self._headers, 12
        )
        self.pre_excavated_depth = utils.parse_measurement_var_as_float(
            self._headers, 13
        )
        self.groundwater_level = utils.parse_measurement_var_as_float(self._headers, 14)
        self.water_depth_offshore_activities = utils.parse_measurement_var_as_float(
            self._headers, 15
        )
        self.end_depth_of_penetration_test = utils.parse_measurement_var_as_float(
            self._headers, 16
        )
        self.stop_criteria = utils.parse_measurement_var_as_float(self._headers, 17)
        self.zero_measurement_cone_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 20)
        )
        self.zero_measurement_cone_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 21)
        )
        self.zero_measurement_friction_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 22)
        )
        self.zero_measurement_friction_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 23)
        )
        self.zero_measurement_ppt_u1_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 24)
        )
        self.zero_measurement_ppt_u1_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 25)
        )
        self.zero_measurement_ppt_u2_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 26)
        )
        self.zero_measurement_ppt_u2_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 27)
        )
        self.zero_measurement_ppt_u3_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 28)
        )
        self.zero_measurement_ppt_u3_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 29)
        )
        self.zero_measurement_inclination_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 30)
        )
        self.zero_measurement_inclination_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 31)
        )
        self.zero_measurement_inclination_ns_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 32)
        )
        self.zero_measurement_inclination_ns_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 33)
        )
        self.zero_measurement_inclination_ew_before_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 34)
        )
        self.zero_measurement_inclination_ew_after_penetration_test = (
            utils.parse_measurement_var_as_float(self._headers, 35)
        )
        self.mileage = utils.parse_measurement_var_as_float(self._headers, 41)

        self.columns_info = self.ColumnsInfo(
            *parse_all_columns_info(self._headers, MAP_QUANTITY_NUMBER_COLUMN_NAME_CPT),
            col_separator=utils.get_column_separator(self._headers),
            rec_separator=utils.get_record_separator(self._headers),
            column_voids=utils.parse_column_void(self._headers),
        )

        self.df = (
            self.parse_data(
                self._data,
                self.columns_info.col_separator,
                self.columns_info.rec_separator,
                self.columns_info.descriptions,
            )
            .lazy()
            .pipe(replace_column_void, self.columns_info.description_to_void_mapping)
            # Remove the rows with null values
            .drop_nulls()
            .with_columns(pl.col("penetrationLength").abs().alias("penetrationLength"))
            .pipe(correct_pre_excavated_depth, self.pre_excavated_depth)
            .pipe(correct_depth_with_inclination, self.columns_info.descriptions)
            .collect()
        )


def correct_pre_excavated_depth(lf: pl.LazyFrame, pre_excavated_depth) -> pl.LazyFrame:
    if pre_excavated_depth is not None and pre_excavated_depth > 0:
        return lf.filter(pl.col("penetrationLength") >= pre_excavated_depth)
    return lf


def correct_depth_with_inclination(
    lf: pl.LazyFrame, columns: List[str]
) -> pl.LazyFrame:
    """
    Return the expression needed to correct depth
    """
    if "depth" in columns:
        return lf.with_columns(pl.col("depth").abs().alias("depth"))
    elif "inclinationResultant" in columns:
        # every different in depth needs to be corrected with the angle
        correction_factor = np.cos(
            np.radians(pl.col("inclinationResultant").cast(pl.Float32).fill_null(0))
        )

        delta_height = pl.col("penetrationLength").diff()
        corrected_depth = correction_factor * delta_height

        return lf.with_columns(
            # this sets the first as depth
            pl.when(pl.arange(0, corrected_depth.len()) == 0)
            .then(pl.col("penetrationLength"))
            .otherwise(corrected_depth)
            .cum_sum()
            .alias("depth")
        )

    return lf



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import List, overload

import polars as pl
from numpy.typing import NDArray


@dataclass
class Location:
    """DataClass that holds the standardized location information"""

    srs_name: str
    x: float
    y: float


class VerticalDatumClass(Enum):
    """DataClass that holds the standardized vertical reference information"""

    Unknown = "-1"
    NAP = "31000"  # Normaal Amsterdams Peil
    BOP = "32000"  # Ostend Level (GEF)
    MSL = "49000"  # Mean Sea Level
    LAT = "00001"  # Lowest Astronomical Tide
    TAW = "32001"  # (GEF)


def assign_multiple_columns(
    df: pl.DataFrame, columns: List[str], partial_df: pl.DataFrame
) -> pl.DataFrame:
    """
    Helper function to assign multiple columns

    :param df: original Dataframe
    :param columns: updated column values
    :param partial_df: updated Dataframe
    :return: DataFrame
    """
    return df.drop(columns).hstack(partial_df[columns])


def kpa_to_mpa(df: pl.DataFrame, columns: List[str]) -> pl.DataFrame:
    """
    Transform from kPa to MPa.

    :param df: DataFrame
    :param columns: columns names
    :return: DataFrame
    """
    return assign_multiple_columns(df, columns, df[columns] * 10**-3)


@overload
def offset_to_depth(ref: float, offset: float) -> float: ...


@overload
def offset_to_depth(ref: NDArray, offset: float) -> NDArray: ...


def offset_to_depth(
    ref: float | None | NDArray, offset: float | None
) -> float | None | NDArray:
    """
    Transform depth with respect to reference level to depth

    :param offset: surface reference level
    :param ref: z
    :return: depth
    """
    if offset is None or ref is None:
        return None
    return -(ref - offset)


@overload
def depth_to_offset(depth: float | None, offset: float | None) -> float | None: ...


@overload
def depth_to_offset(depth: NDArray, offset: float) -> NDArray: ...


def depth_to_offset(
    depth: float | None | NDArray, offset: float | None
) -> float | None | NDArray:
    """
    Transform depth to depth with respect to reference level

    :param offset: surface reference level
    :param depth: z
    :return: depth with respect to reference level
    """
    if offset is None or depth is None:
        return None
    return offset - depth



from __future__ import annotations

import logging
import re
from dataclasses import dataclass
from typing import Dict, List, Tuple, Union

import polars as pl
from gef_file_to_map import gef_to_map

from pygef import exceptions
from pygef.gef import utils

logger = logging.getLogger(__name__)


class _Gef:
    """
    The gef parser is built following the conventional format described in:
    https://publicwiki.deltares.nl/download/attachments/102204318/GEF-CPT.pdf?version=1&modificationDate=1409732008000&api=v2
    """

    def __init__(self, path=None, string=None):
        """
        Base class of gef parser. It switches between the cpt or borehole parser.

        It takes as input either the path to the gef file or the gef file as string.

        Parameters
        ----------
        path: str
            Path to the *.gef file.
        string: str
            String version of the *.gef file.

        """
        self.path = path
        self.df = None
        self.net_surface_area_quotient_of_the_cone_tip = None
        self.pre_excavated_depth = None

        if string is None:
            with open(path, encoding="utf-8", errors="ignore") as f:
                string = f.read()

        self.s = string

        # Use the Rust optimized header parser
        self._data, self._headers = gef_to_map(string)

        self.zid = utils.parse_zid_as_float(self._headers)
        self.height_system = utils.parse_height_system(self._headers)
        self.x = utils.parse_xid_as_float(self._headers)
        self.y = utils.parse_yid_as_float(self._headers)
        self.file_date = utils.parse_file_date(self._headers)
        self.test_id = utils.parse_test_id(self._headers)
        self.type = utils.parse_gef_type(self._headers)

    @dataclass
    class ColumnsInfo:
        """
        Dataclass carrying the #COLUMNINFO headers
        All inputs should be sorted on increasing column-number.
        """

        column_numbers: List[int]
        units: List[str]
        descriptions: List[str]
        column_quantities: List[int]
        col_separator: str
        rec_separator: str
        column_voids: Dict[int, float]
        default_void: float = float(-9999)

        def __post_init__(self):
            self.columns_number = len(self.column_numbers)
            self.column_voids = self.fill_default_column_voids()

            self.validate()

        @property
        def description_to_void_mapping(self) -> Dict[str, float]:
            return {
                desc: self.column_voids[num]
                for num, desc in zip(self.column_numbers, self.descriptions)
            }

        def fill_default_column_voids(self) -> Dict[int, float]:
            default_column_voids = {
                num: self.default_void for num in self.column_numbers
            }
            default_column_voids.update(self.column_voids)
            return default_column_voids

        def validate(self):
            if not len(self.column_numbers) == self.column_numbers[-1]:
                raise exceptions.ParseGefError(
                    "One or more #COLUMNINFO headers are missing: We counted "
                    f"{len(self.column_numbers)} #COLUMNINFO headers, but the highest column index "
                    f"is {self.column_numbers[-1]}"
                )

            if not self.column_numbers == list(range(1, self.column_numbers[-1] + 1)):
                raise exceptions.ParseGefError(
                    "One or more #COLUMNINFO headers have duplicates in the .gef file."
                )

    @staticmethod
    def parse_data(
        data_s: str,
        col_separator: str,
        rec_separator: str,
        column_names: List[str],
    ) -> pl.DataFrame:
        """
        Parses all data in the data string, returns a pl.DataFrame

        Parameters
        ----------
        data_s: str
            A string with the measurement data, in Delimiter-separated format.
        col_separator: str
            The character that separates the columns
        rec_separator: str
            The character that separates the records/rows
        column_names: List[str]
            List of column names

        Returns
        -------
        df: pl.DataFrame
            The DataFrame with measurement data
        """

        # Remove all horizontal whitespace characters around the column separator
        new_data = re.sub(
            rf"[^\S\r\n]*{re.escape(col_separator)}[^\S\r\n]*",
            col_separator,
            data_s,
        )

        # Split string by record separator into lines
        # Remove all whitespaces and column separators at the beginning and end of lines
        # Also remove the last trailing line
        regex = rf"[\s{re.escape(col_separator)}]+"
        new_data = "\n".join(
            re.sub(f"{regex}$", "", re.sub(f"^{regex}", "", line))
            for line in new_data.split(rec_separator)
        ).rstrip()

        return pl.read_csv(
            new_data.encode(),
            separator=col_separator,
            new_columns=column_names,
            has_header=False,
            columns=list(range(0, len(column_names))),
        )


def replace_column_void(
    lf: pl.LazyFrame, col_name_to_void_mapping: Dict[str, float]
) -> pl.LazyFrame:
    return (
        # Get all values matching column_void and change them to null
        # Interpolate all null values
        lf.select(
            [
                pl.when(pl.col(col) == pl.lit(col_name_to_void_mapping[col]))
                .then(None)
                .otherwise(pl.col(col))
                .interpolate()
                .name.keep()
                for col in lf.collect_schema().names()
            ]
        )
    )


def parse_all_columns_info_from_dict(
    headers: dict, quantity_dict: dict
) -> List[Tuple[int, str, str, int]]:
    """
    Function that parses and returns all COLUMNINFO header data from a dictionary.

    :param headers:(str) String of headers.
    :param quantity_dict: (dict) Dictionary that maps quanity numbers to descriptions
    :return: (List[Tuple[int, str, str, int]]) List of Tuples with the COLUMNINFO header values
    """

    columns_info: List[Tuple[int, str, str, int]] = []

    if "COLUMNINFO" in headers:
        # Loop over all headers to find the right number
        for values in headers["COLUMNINFO"]:
            try:
                col_number = int(values[0])
                quantity = int(values[3])
            except ValueError:
                raise exceptions.ParseGefError(
                    "One or more #COLUMNINFO headers do not have integers in the first "
                    "and/or last positions, which is required."
                )

            unit = values[1]
            description = utils.get_description(quantity, values[2], quantity_dict)

            columns_info.append((col_number, unit, description, quantity))

    return columns_info


def parse_all_columns_info_from_str(
    headers: str, quantity_dict: dict
) -> List[Tuple[int, str, str, int]]:
    """
    Function that parses and returns all COLUMNINFO header data from a string.

    :param headers:(dict) Dictionary of headers.
    :param quantity_dict: (dict) Dictionary that maps quanity numbers to descriptions
    :return: (List[Tuple[int, str, str, int]]) List of Tuples with the COLUMNINFO header values
    """

    columns_info: List[Tuple[int, str, str, int]] = []

    # Find all '#COLUMNINFO= **,' strings first
    for match in re.finditer(
        r"#COLUMNINFO\s*=\s*(\d+)\s*,{1}\s*([^,]+)\s*,{1}([^,]+)\s*,{1}\s*(\d+)",
        headers,
    ):
        column_number = utils.cast_string(int, match.group(1))
        column_quantity = utils.cast_string(int, match.group(4))

        custom_description = match.group(3).strip()
        description = utils.get_description(
            column_quantity, custom_description, quantity_dict
        )

        if column_number is not None and column_quantity is not None:
            columns_info.append(
                (
                    column_number,
                    match.group(2).strip(),
                    description,
                    column_quantity,
                )
            )

    return columns_info


def parse_all_columns_info(
    headers: Union[dict, str], quantity_dict: dict
) -> Tuple[List[int], List[str], List[str], List[int]]:
    """
    Function that parses and returns all COLUMNINFO header data from a string or dictionary.

    :param headers:(Union[Dict,str]) Dictionary or string of headers.
    :param quantity_dict: (dict) Dictionary that maps quanity numbers to descriptions
    :return: Tuple with lists of column_numbers, units, descriptions and column_quantities, sorted by column_numbers
    """

    if isinstance(headers, dict):
        columns_info = parse_all_columns_info_from_dict(headers, quantity_dict)

    else:
        columns_info = parse_all_columns_info_from_str(headers, quantity_dict)

    columns_info.sort()

    column_numbers = []
    units = []
    descriptions = []
    column_quantities = []

    for col in columns_info:
        column_numbers.append(col[0])
        units.append(col[1])
        descriptions.append(col[2])
        column_quantities.append(col[3])

    return column_numbers, units, descriptions, column_quantities



from __future__ import annotations

import io
import os
import re
from pathlib import Path

from lxml import etree

from pygef.bore import BoreData
from pygef.broxml import resolvers
from pygef.broxml.xml_parser import BaseParser, read_xml

# maps keyword argument to:
# xpath: query passed to elementree.find
# resolver: A function that converts the string to the proper datatype
#           Fn(str) -> Any
# el-atrr: Optional: attribute of an element taken before send to resolver
# Version 1 not supported, this is a WIP
BORE_ATTRIBS_V1 = {
    "research_report_date": {
        "xpath": "./researchReportDate/brocom:date",
        "resolver": resolvers.parse_date,
        "el-attr": "text",
    },
    "description_procedure": {
        "xpath": "./isbhrgt:boreholeSampleDescription/bhrgtcom:descriptionProcedure",
        "el-attr": "text",
    },
    "delivered_location": {
        "xpath": "./isbhrgt:deliveredLocation/bhrgtcom:location/gml:Point",
        "resolver": resolvers.parse_gml_location,
    },
    "delivered_vertical_position_offset": {
        "xpath": "./isbhrgt:deliveredVerticalPosition/bhrgtcom:offset",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "delivered_vertical_position_datum": {
        "xpath": "./isbhrgt:deliveredVerticalPosition/bhrgtcom:verticalDatum",
        "resolver": resolvers.lower_text,
        "el-attr": "text",
    },
    "delivered_vertical_position_reference_point": {
        "xpath": "./isbhrgt:deliveredVerticalPosition/bhrgtcom:localVerticalReferencePoint",
        "resolver": resolvers.lower_text,
        "el-attr": "text",
    },
    "bore_rock_reached": {
        "xpath": "./isbhrgt:boring/bhrgtcom:rockReached",
        "resolver": resolvers.parse_bool,
        "el-attr": "text",
    },
    "final_bore_depth": {
        "xpath": "./isbhrgt:boring/bhrgtcom:finalDepthBoring",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "final_sample_depth": {
        "xpath": "./isbhrgt:boring/bhrgtcom:finalDepthSampling",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "bore_hole_completed": {
        "xpath": "./isbhrgt:boring",
        "resolver": resolvers.parse_bool,
        "el-attr": "text",
    },
}

BORE_ATTRIBS_V2 = {
    "bro_id": {"xpath": "brocom:broId"},
    "research_report_date": {
        "xpath": "./researchReportDate/brocom:date",
        "resolver": resolvers.parse_date,
        "el-attr": "text",
    },
    "groundwater_level": {
        "xpath": "./boring/bhrgtcom:groundwaterLevel",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "description_procedure": {
        "xpath": "./boreholeSampleDescription/bhrgtcom:descriptionProcedure",
        "el-attr": "text",
    },
    "delivered_location": {
        "xpath": "./deliveredLocation/bhrgtcom:location/gml:Point",
        "resolver": resolvers.parse_gml_location,
    },
    "standardized_location": {
        "xpath": "./standardizedLocation/brocom:location",
        "resolver": resolvers.parse_gml_location,
    },
    "delivered_vertical_position_offset": {
        "xpath": "./deliveredVerticalPosition/bhrgtcom:offset",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "delivered_vertical_position_datum": {
        "xpath": "./deliveredVerticalPosition/bhrgtcom:verticalDatum",
        "resolver": resolvers.lower_text,
        "el-attr": "text",
    },
    "delivered_vertical_position_reference_point": {
        "xpath": "./deliveredVerticalPosition/bhrgtcom:localVerticalReferencePoint",
        "resolver": resolvers.lower_text,
        "el-attr": "text",
    },
    "bore_rock_reached": {
        "xpath": "./boring/bhrgtcom:rockReached",
        "resolver": resolvers.parse_bool,
        "el-attr": "text",
    },
    "final_bore_depth": {
        "xpath": "./boring/bhrgtcom:finalDepthBoring",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "final_sample_depth": {
        "xpath": "./boring/bhrgtcom:finalDepthSampling",
        "resolver": resolvers.parse_float,
        "el-attr": "text",
    },
    "bore_hole_completed": {
        "xpath": "./boring/bhrgtcom:boreholeCompleted",
        "resolver": resolvers.parse_bool,
        "el-attr": "text",
    },
    "data": {
        "xpath": "./boreholeSampleDescription/bhrgtcom:descriptiveBoreholeLog",
        "resolver": resolvers.process_bore_result,
    },
}


def read_bore(file: io.BytesIO | Path | str) -> list[BoreData]:
    if isinstance(file, str) and not os.path.exists(file):
        root = etree.fromstring(file, parser=BaseParser).getroot()
    else:
        root = etree.parse(file, parser=BaseParser).getroot()
    match = re.compile(r"xsd/.*/(\d\.\d)")
    matched = match.search(root.nsmap["bhrgtcom"])

    if matched is None:
        raise ValueError("could not find the brhtcom version")
    else:
        if 3.0 >= float(matched.group(1)) < 2.0:
            raise ValueError("only bhrgtcom/2.x is supported ")
        return read_xml(root, BoreData, BORE_ATTRIBS_V2, "dispatchDocument")



from __future__ import annotations

import re
from datetime import date, datetime
from typing import Any, List
from warnings import warn

import polars as pl
from lxml import etree

from pygef.common import Location, VerticalDatumClass
from pygef.gef.utils import parse_regex_cast


def lower_text(val: str, **kwargs: dict[Any, Any]) -> str:
    return val.lower()


def parse_float(val: str | None, **kwargs: dict[Any, Any]) -> float | None:
    if isinstance(val, (str, int, float)):
        return float(val)
    return None


def parse_int(val: str | None, **kwargs: dict[Any, Any]) -> int | None:
    if isinstance(val, (str, int, float)):
        return int(val)
    return None


def parse_date(val: str, **kwargs: dict[Any, Any]) -> date:
    return datetime.strptime(val, "%Y-%m-%d").date()


def clean_string(val: str, **kwargs: dict[Any, Any]) -> str:
    if isinstance(val, str):
        return re.sub(r"\W+", "", val)
    return "unknown"


def parse_bool(val: str, **kwargs: dict[Any, Any]) -> bool:
    val = val.lower()
    if val == "ja":
        return True
    if val == "nee" or val == "geen":
        return False
    return bool(val)


def process_bore_result(el: etree.Element, **kwargs: dict[Any, Any]) -> pl.DataFrame:
    namespaces = kwargs["namespaces"]
    upper_boundary = []
    lower_boundary = []
    geotechnical_soil_name_iso: List[str] = []
    geotechnical_soil_name_nen: List[str] = []
    color: List[str] = []
    dispersed_inhomogenity: List[bool | None] = []
    organic_matter_content_class: List[str | None] = []
    sand_median_class: List[str | None] = []
    for layer in el.iterfind("bhrgtcom:layer", namespaces=namespaces):
        upper_boundary.append(
            float(layer.find("bhrgtcom:upperBoundary", namespaces=namespaces).text)
        )
        lower_boundary.append(
            float(layer.find("bhrgtcom:lowerBoundary", namespaces=namespaces).text)
        )
        try:
            geotechnical_soil_name_iso.append(
                clean_string(
                    layer.find(
                        "bhrgtcom:soil/bhrgtcom:geotechnicalSoilName",
                        namespaces=namespaces,
                    ).text
                )
            )
        except AttributeError:
            geotechnical_soil_name_iso.append("niet gedefinieerd")
        try:
            geotechnical_soil_name_nen.append(
                clean_string(
                    layer.find(
                        "bhrgtcom:soil/bhrgtcom:soilNameNEN5104",
                        namespaces=namespaces,
                    ).text
                )
            )
        except AttributeError:
            geotechnical_soil_name_nen.append("niet gedefinieerd")
        try:
            color.append(
                clean_string(
                    layer.find(
                        "bhrgtcom:soil/bhrgtcom:colour", namespaces=namespaces
                    ).text
                )
            )
        except AttributeError:
            color.append("onbekend")

        try:
            dispersed_inhomogenity.append(
                parse_bool(
                    layer.find(
                        "bhrgtcom:soil/bhrgtcom:dispersedInhomogeneity",
                        namespaces=namespaces,
                    ).text
                )
            )
        except AttributeError:
            dispersed_inhomogenity.append(None)
        try:
            organic_matter_content_class.append(
                clean_string(
                    layer.find(
                        "bhrgtcom:soil/bhrgtcom:organicMatterContentClass",
                        namespaces=namespaces,
                    ).text
                )
            )
        except AttributeError:
            organic_matter_content_class.append(None)
        try:
            sand_median_class.append(
                clean_string(
                    layer.find(
                        "bhrgtcom:soil/bhrgtcom:sandMedianClass", namespaces=namespaces
                    ).text
                )
            )
        except AttributeError:
            sand_median_class.append(None)

    # merge NEN and ISO names
    geotechnical_soil_name = [
        word if word != "unknown" else geotechnical_soil_name_nen[i]
        for i, word in enumerate(geotechnical_soil_name_iso)
    ]
    return pl.DataFrame(
        {
            "upperBoundary": upper_boundary,
            "lowerBoundary": lower_boundary,
            "geotechnicalSoilName": geotechnical_soil_name,
            "color": color,
            "dispersedInhomogeneity": dispersed_inhomogenity,
            "organicMatterContentClass": organic_matter_content_class,
            "sandMedianClass": sand_median_class,
        }
    )


def process_cpt_result(el: etree.Element, **kwargs: dict[Any, Any]) -> pl.DataFrame:
    """
    Parse the cpt data into a `DataFrame`

    Parameters
    ----------
    el
        conePenetrometerSurvey
    kwargs
        namespaces.
    """
    namespaces = kwargs["namespaces"]

    prefix = "./cptcommon:conePenetrationTest/cptcommon:cptResult"

    text_enc = el.find(f"{prefix}/swe:encoding/swe:TextEncoding", namespaces=namespaces)
    decimal_sep = text_enc.attrib["decimalSeparator"]
    if decimal_sep != ".":
        warn(
            f"Found a '{decimal_sep}' as decimal separator, this may lead to parsing errors."
        )

    delimiter = text_enc.attrib["tokenSeparator"]
    new_line_char = text_enc.attrib["blockSeparator"]

    columns = []
    selection = []

    i = 0
    for param in el.find(
        "./cptcommon:parameters", namespaces=namespaces
    ).iterchildren():
        name = param.tag.split("}")[1]
        if parse_bool(param.text):
            columns.append(name)
            # we select the columns by index
            # this prevents materializing invalid columns
            selection.append(i)
        i += 1

    # we strip the data because there is leading and trailing whitespace.
    data = el.find(f"{prefix}/cptcommon:values", namespaces=namespaces).text.strip()
    return pl.read_csv(
        data.encode(),
        new_columns=columns,
        columns=selection,
        has_header=False,
        schema_overrides=[pl.Float64] * len(columns),
        separator=delimiter,
        eol_char=new_line_char,
        ignore_errors=True,
        null_values="-999999",
    ).drop_nulls("coneResistance")


def parse_gml_location(el: etree.Element, **kwargs: dict[Any, Any]) -> Location:
    """Resolver for standardizedLocation/brocom:location"""
    srs_name = el.attrib["srsName"]
    pos = next(el.iterfind("./gml:pos", namespaces=kwargs["namespaces"])).text
    (x, y) = parse_position(pos)
    return Location(srs_name=srs_name, x=x, y=y)


def parse_quality_class(val: str, **kwargs: dict[Any, Any]) -> int | None:
    val = val.lower().replace(" ", "")

    return parse_regex_cast(r"^.*?(klasse|class|kl.).*?(\d{1})", val, int, 2)


def parse_local_vertical_reference_point_class(
    val: str, **kwargs: dict[Any, Any]
) -> VerticalDatumClass:
    val = val.lower().strip()
    if val == "nap":
        return VerticalDatumClass.NAP
    if val == "msl":
        return VerticalDatumClass.MSL
    if val == "lat":
        return VerticalDatumClass.LAT
    warn(f"vertical datum class '{val}' is unknown")
    return VerticalDatumClass.Unknown


def parse_position(pos: str) -> tuple[float, float]:
    """
    Parse a position tuple

    Parameters
    ----------
    pos
        Any of {'x y', 'x,y', 'x;y'}
        where x and y are parsable by float

    Returns
    -------

    """

    if " " in pos:
        splitter = " "
    elif "," in pos:
        splitter = ","
    elif ";" in pos:
        splitter = ";"
    else:
        raise ValueError(f"pygef does not know how to parse '{pos}' position")
    parts = pos.split(splitter)
    return float(parts[0]), float(parts[1])



from __future__ import annotations

import copy
import pprint
from dataclasses import dataclass, field
from datetime import date
from typing import Any

import polars as pl

from pygef.broxml.mapping import MAPPING_PARAMETERS
from pygef.common import Location, depth_to_offset


@dataclass(frozen=True)
class BoreData:
    """
    The Bore dataclass holds the information from the BHRgt object.

    Attributes:
        bro_id (str | None): BRO ID of the BHRgt.
        alias  (str | None): Alias of the CPT.
        research_report_date (date): research report date
        delivered_location (Location): delivered location in EPSG:28992 - RD new
        groundwater_level (float | None): groundwater level
        standardized_location (Location | None): standardized location in EPSG:4326 - WGS 84
        delivered_vertical_position_offset (float | None): delivered vertical position offset
        delivered_vertical_position_datum (str): research delivered vertical position datum
        delivered_vertical_position_reference_point (str): delivered vertical position reference point
        bore_rock_reached (bool): bore rock reached
        final_bore_depth (float): final bore depth
        data (pl.DataFrame): DataFrame
            columns:

                - upperBoundary [m]
                - lowerBoundary [m]
                - geotechnicalSoilName
                - soilDistribution

            optional columns:

                - upperBoundaryOffset [m wrt offset]
                    see delivered_vertical_position_datum for offset
                - lowerBoundaryOffset [m wrt offset]
                    see delivered_vertical_position_datum for offset
                - lutumPercentage [%]
                - siltPercentage [%]
                - sandPercentage [%]
                - gravelPercentage [%]
                - organicMatterPercentage [%]
                - sandMedianClass
                - gravelMedianClass
                - geotechnicalSoilCode
                - color
                - remarks
                - dispersedInhomogeneity
                - organicMatterContentClass
    """

    # dispatch_document bhrgt
    bro_id: str | None
    research_report_date: date
    description_procedure: str
    delivered_location: Location
    groundwater_level: float | None
    standardized_location: Location | None
    delivered_vertical_position_offset: float | None
    delivered_vertical_position_datum: str
    delivered_vertical_position_reference_point: str
    bore_rock_reached: bool
    final_bore_depth: float
    final_sample_depth: float | None
    bore_hole_completed: bool
    data: pl.DataFrame

    alias: str | None = field(default=None)

    def __post_init__(self):
        # post-processing of the data
        tbl = MAPPING_PARAMETERS.dist_table().lazy()
        df = (
            self.data.lazy()
            .pipe(
                _calculate_depth_with_respect_to_offset,
                self.delivered_vertical_position_offset,
            )
            .join(tbl, on="geotechnicalSoilName", how="left")
            .sort("upperBoundary", descending=False, nulls_last=False)
            .collect()
        )
        # bypass FrozenInstanceError
        object.__setattr__(self, "data", df)

    @property
    def columns(self) -> list[str]:
        """Columns names for the DataFrame"""
        return self.data.columns

    @property
    def groundwater_level_offset(self) -> float | None:
        """groundwater level wrt offset"""
        return depth_to_offset(
            self.groundwater_level, offset=self.delivered_vertical_position_offset
        )

    @property
    def final_sample_depth_offset(self) -> float | None:
        """final sample depth wrt offset"""
        return depth_to_offset(
            self.final_sample_depth, offset=self.delivered_vertical_position_offset
        )

    @property
    def final_depth_offset(self) -> float | None:
        """final depth wrt offset"""
        return depth_to_offset(
            self.final_depth_offset, offset=self.delivered_vertical_position_offset
        )

    def __str__(self):
        return f"BoreData: {self.display_attributes()}"

    def attributes(self) -> dict[str, Any]:
        """
        Get the attributes
        """
        attribs = copy.copy(self.__dict__)
        attribs["data"] = attribs["data"].shape
        return attribs

    def display_attributes(self) -> str:
        """
        Get pretty formatted string representation of `CPTData.attributes``
        """
        return pprint.pformat(self.attributes())


def _calculate_depth_with_respect_to_offset(
    lf: pl.LazyFrame, offset: float | None
) -> pl.LazyFrame:
    """post-process function for CPT data creates a new column with the elevation with respect to offset"""
    if offset is None:
        return lf

    return lf.with_columns(
        (offset - pl.col("upperBoundary")).alias("upperBoundaryOffset"),
        (offset - pl.col("lowerBoundary")).alias("lowerBoundaryOffset"),
    )



from datetime import datetime

import pytest
from lxml.etree import XMLSyntaxError

from pygef import read_cpt
from pygef.common import Location, VerticalDatumClass
from pygef.cpt import CPTData


def test_engine(cpt_gef_1) -> None:
    # read test with force incorrect engine
    with pytest.raises(XMLSyntaxError):
        read_cpt(cpt_gef_1, engine="xml")
    # read test with force engine
    gef = read_cpt(cpt_gef_1, engine="gef")
    isinstance(gef, CPTData)
    # read test with auto engine
    gef = read_cpt(cpt_gef_1, engine="auto")
    isinstance(gef, CPTData)


@pytest.mark.parametrize("_type", ["string", "path", "byte"])
def test_gef_to_cpt_data(_type, cpt_gef_1, cpt_gef_1_bytes, cpt_gef_1_string) -> None:
    _format = {
        "string": cpt_gef_1_string,
        "path": cpt_gef_1,
        "byte": cpt_gef_1_bytes,
    }

    cpt_data = read_cpt(_format[_type])

    assert cpt_data.attributes() == {
        "bro_id": None,
        "alias": "CPTU17.8 + 83BITE",
        "cone_diameter": None,
        "cone_surface_area": 1000.0,
        "cone_surface_quotient": 0.8,
        "cone_to_friction_sleeve_distance": 80.0,
        "cone_to_friction_sleeve_surface_area": None,
        "cone_to_friction_sleeve_surface_quotient": 1.0,
        "cpt_description": "",
        "cpt_standard": None,
        "groundwater_level": None,
        "cpt_type": 4.0,
        "data": (999, 12),
        "delivered_vertical_position_datum": VerticalDatumClass("31000"),
        "delivered_vertical_position_offset": -0.09,
        "delivered_vertical_position_reference_point": "unknown",
        "dissipationtest_performed": None,
        "final_depth": 20.0,
        "predrilled_depth": 0.0,
        "quality_class": int(2),
        "research_report_date": datetime(2019, 2, 13).date(),
        "standardized_location": None,
        "delivered_location": Location(
            srs_name="urn:ogc:def:crs:EPSG::28992", x=79578.38, y=424838.97
        ),
        "zlm_cone_resistance_after": -0.245,
        "zlm_cone_resistance_before": -0.257,
        "zlm_inclination_ew_after": None,
        "zlm_inclination_ew_before": None,
        "zlm_inclination_ns_after": None,
        "zlm_inclination_ns_before": None,
        "zlm_inclination_resultant_after": None,
        "zlm_inclination_resultant_before": None,
        "zlm_local_friction_after": -0.016,
        "zlm_local_friction_before": -0.015,
        "zlm_pore_pressure_u1_after": None,
        "zlm_pore_pressure_u1_before": None,
        "zlm_pore_pressure_u2_after": -0.013,
        "zlm_pore_pressure_u2_before": -0.028,
        "zlm_pore_pressure_u3_after": None,
        "zlm_pore_pressure_u3_before": None,
    }



from __future__ import annotations

from functools import lru_cache


class _MappingParameters:
    @lru_cache(1)
    def code_to_text(self) -> dict[str, str]:
        return {
            "DO": "dark ",
            "LI": "light ",
            "TBL": "blue-",
            "TBR": "brown-",
            "TGE": "yellow-",
            "TGN": "green-",
            "TGR": "gray-",
            "TOL": "olive-",
            "TOR": "orange-",
            "TPA": "violet-",
            "TRO": "red-",
            "TWI": "white-",
            "TRZ": "pink-",
            "TZW": "black-",
            "BL": "blue ",
            "BR": "brown ",
            "GE": "yellow ",
            "GN": "green ",
            "GR": "gray ",
            "OL": "olive ",
            "OR": "orange ",
            "PA": "violet ",
            "RO": "red ",
            "WI": "white ",
            "RZ": "pink ",
            "ZW": "black ",
            "ZUF": "uiterst fijn ",
            "ZZF": "zeer fijn ",
            "ZMF": "matig fijn ",
            "ZMG": "matig grof ",
            "ZZG": "zeer grof ",
            "ZUG": "uiterst grof ",
            "SZK": "zeer kleine spreiding ",
            "SMK": "matig kleine spreiding|",
            "SMG": "matig grote spreiding ",
            "SZG": "zeer grote spreiding ",
            "STW": "tweetoppige spreiding ",
            "ZZH": "sterk hoekig ",
            "ZHK": "hoekig ",
            "ZMH": "matig hoekig ",
            "ZMA": "afgerond ",
            "ZSA": "sterk afgerond ",
            "GFN": "fijn grind ",
            "GMG": "matig grof grind ",
            "GZG": "zeer grof grind ",
            "FN1": "spoor fijn grind (<1%) ",
            "FN2": "weinig fijn grind (1-25%) ",
            "FN3": "veel fijn grind (25-50%) ",
            "FN4": "zeer veel fijn grind (50-75%) ",
            "FN5": "uiterst veel fijn grind (>75%) ",
            "MG1": "spoor matig grof grind (<1%) ",
            "MG2": "weinig matig grof grind (1-25%) ",
            "MG3": "veel matig grof grind (25-50%) ",
            "MG4": "zeer veel matig grof grind (50-75%) ",
            "MG5": "uiterst veel matig grof grind(>75%) ",
            "GG1": "spoor zeer grof grind (<1%) ",
            "GG2": "weinig zeer grof grind (1-25%) ",
            "GG3": "veel zeer grof grind (25-50%) ",
            "GG4": "zeer veel zeer grof grind (50-75%) ",
            "GG5": "uiterst veel zeer grof grind (>75%) ",
            "AV1": "zwak amorf ",
            "AV2": "matig amorf ",
            "AV3": "sterk amorf ",
            "BSV": "bosveen ",
            "HEV": "heideveen ",
            "MOV": "mosveen ",
            "RIV": "rietveen ",
            "SZV": "Scheuchzeriaveen ",
            "VMV": "veenmosveen ",
            "WOV": "wollegrasveen ",
            "ZEV": "zeggeveen ",
            "KZSL": "klei zeer slap ",
            "KSLA": "klei slap ",
            "KMSL": "klei matig slap ",
            "KMST": "klei stevig ",
            "KZST": "klei zeer stevig ",
            "KHRD": "klei hard ",
            "KZHR": "klei zeer hard ",
            "LZSL": "leem zeer slap ",
            "LSLA": "leem slap ",
            "LMSL": "leem matig slap ",
            "LMST": "leem matig stevig ",
            "LSTV": "leem stevig ",
            "LZST": "leem zeer stevig ",
            "LHRD": "leem hard ",
            "LZHR": "leem zeer hard ",
            "VZSL": "veen zeer slap ",
            "VSLA": "veen slap ",
            "VMSL": "veen matig slap ",
            "VMST": "veen matig stevig ",
            "VSTV": "stevig ",
            "LOS": "los gepakt ",
            "NOR": "normaal gepakt ",
            "VAS": "vast gepakt ",
            "VGZZ": "zeer zacht ",
            "VGZA": "zacht ",
            "VGMZ": "matig zacht ",
            "VGMH": "matig hard ",
            "VGHA": "hard ",
            "VGZH": "zeer hard ",
            "VGEH": "extreem hard ",
            "SCH0": "geen schelpmateriaal 0% ",
            "SCH1": "spoor schelpmateriaal <1% ",
            "SCH2": "weinig schelpmateriaal (1-10%) ",
            "SCH3": "veel schelpmateriaal (10-30%) ",
            "CA1": "kalkloos ",
            "CA2": "kalkarm ",
            "CA3": "kalkrijk ",
            "GC0": "geen glauconiet 0% ",
            "GC1": "spoor glauconiet <1% ",
            "GC2": "weinig glauconiet (1-10%) ",
            "GC3": "veel glauconiet (10-30%) ",
            "GC4": "zeer veel glauconiet(30-50%) ",
            "GC5": "uiterst veel glauconiet (>50%) ",
            "BST1": "spoor baksteen ",
            "BST2": "weinig baksteen ",
            "BST3": "veel baksteen ",
            "PUR1": "spoor puinresten ",
            "PUR2": "weinig puinresten ",
            "PUR3": "veel puinresten ",
            "SIN1": "spoor sintels ",
            "SIN2": "weinig sintels ",
            "SIN3": "veel sintels ",
            "STO1": "spoor stortsteen ",
            "STO2": "weinig stortsteen ",
            "STO3": "veel stortsteen ",
            "VUI1": "spoor vuilnis ",
            "VUI2": "weinig vuilnis ",
            "VUI3": "veel vuilnis ",
            "AF": "afval ",
            "AS": "asfalt ",
            "BE": "beton ",
            "BI": "bitumen ",
            "BT": "ballast ",
            "BST": "baksteen ",
            "GI": "gips ",
            "GA": "glas ",
            "HK": "houtskool ",
            "HU": "huisvuil ",
            "KA": "kalk ",
            "KG": "kolengruis ",
            "KO": "kolen ",
            "KT": "krijt ",
            "ME": "metaal ",
            "MI": "mijnsteen ",
            "OE": "oer ",
            "PL": "planten ",
            "PU": "puin ",
            "SI": "sintels ",
            "SL": "slakken ",
            "WO": "wortels ",
            "YZ": "ijzer ",
            "GL": "gley ",
            "RT": "roest ",
            "SE": "silex ",
            "BIO": "bioturbatie ",
            "DWO": "doorworteling ",
            "GCM": "cm-gelaagdheid ",
            "GDM": "dm-gelaagdheid ",
            "GDU": "dubbeltjes-gelaagdheid ",
            "GMM": "mm-gelaagdheid ",
            "GRG": "graafgangen ",
            "GSC": "scheve gelaagdheid ",
            "GSP": "spekkoek-gelaagdheid ",
            "HOM": "homogeen ",
            "GE1": "zwak gelaagd ",
            "GE2": "weinig gelaagd ",
            "GE3": "sterk gelaagd ",
            "GEX": "gelaagd ",
            "STGL": "met grindlagen ",
            "STKL": "met kleilagen ",
            "STLL": "met leemlagen ",
            "STSL": "met stenenlagen ",
            "STVL": "met veenlagen ",
            "STZL": "met zandlagen ",
            "STBR": "met bruinkoollagen ",
            "STDE": "met detrituslagen ",
            "STGY": "met gyttjalagen ",
            "STSC": "met schelpenlagen ",
            "ANT": "Antropogeen ",
            "BOO": "Boomse klei ",
            "DEZ": "dekzand ",
            "KEL": "keileem ",
            "LSS": "loess ",
            "POK": "potklei ",
            "WAR": "warven ",
            "DR": "Formatie van Drente ",
            "EC": "Formatie van Echteld ",
            "KR": "Formatie van Kreftenheye ",
            "NA": "Formatie van Naaldwijk ",
            "NI": "Formatie van Nieuwkoop ",
            "TW": "Formatie van Twente ",
            "WA": "Formatie van Waalre ",
        }

    @lru_cache(1)
    def dino_to_bro(self, s: str) -> str:
        # TODO update soil_code from NEN 5104 -> NEN-EN-ISO 14688-1:2019+NEN 8990:2020
        main = s[0]

        if len(s) == 2:
            try:
                sub = s[1]
                subf = 0
            except ValueError:
                return "niet gedefinieerd"
        elif len(s) > 2:
            try:
                sub = s[1]
                subf = int(s[2])
            except ValueError:
                return "niet gedefinieerd"
        else:
            if main == "G":
                return "siltigGrind"
            if main == "Z":
                return "zand"
            if main == "L":
                return "zwakZandigSilt"
            if main == "K":
                return "klei"
            if main == "V":
                return "veen"

        # sand
        if main == "Z":
            if sub in "sk":
                if subf > 2:
                    return "sterkZandigSilt"
                else:
                    return "zwakZandigSilt"
            elif sub == "v":
                return "sterkZandigVeen"
            elif sub == "g":
                return "zand"
        # clay
        elif main == "K":
            if sub in "sz":
                if subf > 2:
                    return "sterkZandigeKlei"
                else:
                    return "zwakZandigeKlei"
            elif sub == "g":
                return "sterkZandigeKlei"
            elif sub == "v":
                return "organischKlei"
        elif main == "V":
            return "veen"
        elif main == "L":
            if sub == "z":
                if subf > 2:
                    return "sterkZandigSilt"
                else:
                    return "zwakZandigSilt"
            elif sub == "k":
                if subf > 2:
                    return "zwakZandigeKlei"
                else:
                    return "sterkZandigKlei"
            elif sub == "v":
                if subf > 1:
                    return "zwakZandigVeen"
                else:
                    return "sterkZandigVeen"
            elif sub == "g":
                return "zwakZandigSiltMetGrind"
        elif main == "G":
            if sub in "sz":
                if subf < 2:
                    return "zwakZandigGrind"
                else:
                    return "siltigGrind"
            else:
                return "grind"

        return "niet gedefinieerd"


MAPPING_PARAMETERS = _MappingParameters()
COLUMN_NAMES_BORE = [
    "upperBoundary",  # 1
    "lowerBoundary",  # 2
    "lutumPercentage",  # 3
    "siltPercentage",  # 4
    "sandPercentage",  # 5
    "gravelPercentage",  # 6
    "organicMatterPercentage",  # 7
    "sandMedianClass",  # 8
    "gravelMedianClass",  # 9
]
MAP_QUANTITY_NUMBER_COLUMN_NAME_BORE = dict(enumerate(COLUMN_NAMES_BORE, 1))
MAP_QUANTITY_NUMBER_COLUMN_NAME_CPT = {
    1: "penetrationLength",
    2: "coneResistance",
    3: "localFriction",
    4: "frictionRatio",
    5: "porePressureU1",
    6: "porePressureU2",
    7: "porePressureU3",
    8: "inclinationResultant",
    9: "inclinationNS",
    10: "inclinationEW",
    11: "depth",
    12: "elapsedTime",
    13: "correctedConeResistance",
    14: "netConeResistance",
    15: "poreRatio",
    16: "coneResistanceRatio",
    17: "soilDensity",
    18: "porePressure",
    19: "verticalPorePressureTotal",
    20: "verticalPorePressureEffective",
    21: "inclinationX",
    22: "inclinationY",
    23: "electricalConductivity",
    31: "magneticFieldStrengthX",
    32: "magneticFieldStrengthY",
    33: "magneticFieldStrengthZ",
    34: "magneticFieldStrengthTotal",
    35: "magneticInclination",
    36: "magneticDeclination",
}



from __future__ import annotations

from typing import List, Tuple

import matplotlib.patches as mpatches
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
import numpy as np
from matplotlib.gridspec import GridSpec

from pygef.bore import BoreData
from pygef.cpt import CPTData

try:
    import contextily as ctx
except ImportError:
    ctx = None

FigureSize = (8, 12)
TickLoc = plticker.MultipleLocator(base=0.5)


def plot_cpt(
    data: CPTData,
    ax: plt.Axes | None = None,
    dpi: int = 100,
    use_offset: bool = False,
) -> Tuple[plt.Axes, plt.Axes, plt.Axes]:
    """
    Create a plot with three axes.
        - cone_resistance [MPa]
        - friction_ratio [%] (twiny)
        - friction [MPa] (twiny)

    :param data: (BoreData) Bore data object
    :param ax: (Axes, optional) Axes object used to add axes
    :param dpi: (int, optional)  Default is 100 Resolution of the figure.
    :param use_offset: (bool, optional)  Default is False Plot depth with respect to offset.
    """
    df = data.data
    if use_offset:
        yname = "depthOffset"
        label_name = "depth" if "depth" in data.columns else "penetrationLength"
    else:
        yname = "depth" if "depth" in data.columns else "penetrationLength"
        label_name = yname

    if ax is None:
        fig = plt.figure(figsize=FigureSize, dpi=dpi, layout="tight")
        ax = fig.subplots(1, 1)

    p1 = None
    p11 = None
    p12 = None
    ax2 = ax.twiny()
    p2 = None
    ax3 = ax.twiny()
    p3 = None
    p4 = None

    axes = (ax, ax2, ax3)

    # set the properties on the coneResistance axes
    ax.xaxis.set_ticks_position("top")
    ax.xaxis.set_label_position("top")
    ax.spines["top"].set_position(("axes", 1))
    ax.set_xlim(0, 40)
    ax.set_xlabel("$q_c$ [MPa]")
    ax.xaxis.label.set_color("#2d2e87")
    if not use_offset:
        ax.invert_yaxis()

    # set axis label
    if use_offset:
        ax.set_ylabel(f"{label_name} [m w.r.t. vertical position offset]")
    else:
        ax.set_ylabel(f"{label_name} [m]")

    # create grid
    major_ticks = np.arange(0, 41, 5)
    minor_ticks = np.arange(0, 41, 2.5)
    ax.set_xticks(major_ticks)
    ax.set_xticks(minor_ticks, minor=True)
    ax.grid(which="both")
    ax.grid(which="minor", alpha=0.2)
    ax.grid(which="major", alpha=0.5)
    ax.yaxis.set_major_locator(TickLoc)

    # set the properties on the localFriction axes
    ax2.xaxis.set_ticks_position("top")
    ax2.xaxis.set_label_position("top")
    ax2.spines["top"].set_position(("axes", 1.05))
    ax2.set_xlabel("$f_s$ [MPa]")
    ax2.set_xlim(0, 0.8)
    ax2.xaxis.label.set_color("#e04913")

    # set the properties on the frictionRatio axes
    ax3.xaxis.set_ticks_position("top")
    ax3.xaxis.set_label_position("top")
    ax3.spines["top"].set_position(("axes", 1.1))
    ax3.set_xlabel("$R_f$ [%]")
    ax3.set_xlim(0, 16)
    ax3.invert_xaxis()
    ax3.xaxis.label.set_color("tab:gray")

    # add data to figure
    # plot coneResistance
    if "coneResistance" in data.columns:
        (p1,) = ax.plot(
            df["coneResistance"], df[yname], color="#2d2e87", label="coneResistance"
        )
    # plot localFriction
    if "localFriction" in df.columns:
        (p2,) = ax2.plot(
            df["localFriction"], df[yname], color="#e04913", label="localFriction"
        )

    # plot measured frictionRatio
    if "frictionRatio" in df.columns:
        (p3,) = ax3.plot(
            df["frictionRatio"],
            df[yname],
            label="frictionRatio measured",
            color="tab:gray",
        )

    # plot computed frictionRatio
    if "frictionRatioComputed" in df.columns:
        (p4,) = ax3.plot(
            df["frictionRatioComputed"],
            df[yname],
            label="frictionRatio computed",
            ls="dashed",
            color="tab:gray",
        )
    # add hlines
    if data.groundwater_level is not None:
        value = data.groundwater_level_offset if use_offset else data.groundwater_level
        p11 = ax.axhline(
            value,
            label=f"groundwater level: {value:.2f}",
            color="tab:blue",
            ls="dashed",
        )
    if data.predrilled_depth is not None:
        value = data.predrilled_depth_offset if use_offset else data.predrilled_depth
        p12 = ax.axhline(
            value,
            label=f"predrilled depth: {value:.2f}",
            color="tab:brown",
            ls="dashed",
        )

    # add legend
    ax.legend(
        loc="upper center",
        title=f"CPT: {data.alias if data.bro_id is None else data.bro_id}",
        handles=[i for i in [p1, p2, p3, p4, p11, p12] if i is not None],
    )
    return axes


def plot_bore(
    data: BoreData,
    ax: plt.Axes | None = None,
    dpi: int = 100,
    legend: bool = True,
    use_offset: bool = False,
) -> plt.Axes:
    """
    Create a plot with one axes.
        - soil distribution

    :param data: Bore data object
    :param ax: Axes object
    :param dpi: Default is 100 Resolution of the figure.
    :param legend: Default is True Add legend to the figure.
    :param use_offset: (bool, optional)  Default is False Plot depth with respect to offset.
    """
    if use_offset:
        yupper = "upperBoundaryOffset"
        ylower = "lowerBoundaryOffset"
    else:
        yupper = "upperBoundary"
        ylower = "lowerBoundary"

    if ax is None:
        fig = plt.figure(figsize=FigureSize, dpi=dpi, layout="tight")
        ax = fig.subplots(1, 1)

    df = data.data
    soil_dist = np.stack(df["soilDistribution"].to_numpy())
    cum_soil_dist = soil_dist.cumsum(axis=1)

    # peat, clay, silt, sand, gravel, rocks
    legend_colors = [
        "#a76b29",
        "#578E57",
        "#0078C1",
        "#DBAD4B",
        "#708090",
        "#59626b",
    ]
    legend_names = ["peat", "clay", "silt", "sand", "gravel", "rocks"]
    if not use_offset:
        ax.invert_yaxis()
    # set axis label
    if use_offset:
        ax.set_ylabel("depth [m w.r.t. vertical position offset]")
    else:
        ax.set_ylabel("depth [m]")
    ax.set_xlabel("cumulative soil fraction [-]")
    ax.yaxis.set_major_locator(TickLoc)

    for i in range(5):
        ax.fill_betweenx(
            np.array(list(zip(df[yupper], df[ylower]))).flatten(),
            np.zeros(df.shape[0] * 2),
            np.repeat(cum_soil_dist[:, -(i + 1)], 2),
            color=legend_colors[i],
        )

    # add the name of the soil to the plot
    for row in df.rows(named=True):
        y = (row[ylower] - row[yupper]) / 2 + row[yupper]
        ax.annotate(text=row["geotechnicalSoilName"], xy=(0.25, y))

    # add legend
    if legend:
        patches = [
            mpatches.Patch(color=color, label=key)
            for key, color in zip(legend_names, legend_colors)
        ]
        ax.legend(
            handles=patches,
            bbox_to_anchor=(1, 1),
            loc="upper left",
            title=f"BHRgt: {data.alias if data.bro_id is None else data.bro_id}",
        )

    return ax


def plot_merge(
    bore_data: BoreData, cpt_data: CPTData, dpi: int = 100, basemap: bool = True
) -> Tuple[plt.Figure, List[plt.Axes]]:
    """
    Create one plot that combines the CPT and BHRgt data.

    plot with three axes.
        - cone_resistance [Mpa]
            - friction_ratio [%] (twiny)
            - friction [MPa] (twiny)
        - soil distribution
        - map of the Bore and CPT location

    :param bore_data: Bore data object
    :param cpt_data: CPT data object
    :param dpi: Default is 100 Resolution of the figure.
    :param basemap: Default is True Add a basemap to the map with contextily
    """

    # init subplots
    fig = plt.figure(figsize=FigureSize, dpi=dpi, layout="constrained")
    gs = GridSpec(2, 2, figure=fig, width_ratios=[2, 1], height_ratios=[1, 0.5])
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1], sharey=ax1)
    ax3 = fig.add_subplot(gs[2:])
    axes = [ax1, ax2, ax3]

    # fill axes and update ylabels
    _ = plot_cpt(cpt_data, ax1, use_offset=True)
    _ = plot_bore(bore_data, ax2, legend=False, use_offset=True)
    ax2.set_ylabel("")

    # plot BHRgt location
    ax3.scatter(
        bore_data.delivered_location.x,
        bore_data.delivered_location.y,
        marker="s",
        color="tab:brown",
        label=f"BHRgt: {bore_data.bro_id}",
    )
    # plot cpt location
    ax3.scatter(
        cpt_data.delivered_location.x,
        cpt_data.delivered_location.y,
        marker="v",
        color="tab:green",
        label=f"CPT: {cpt_data.bro_id}",
    )
    # make sure there is no offset in the axis
    ax3.ticklabel_format(useOffset=False, style="plain")
    # set axis to equal
    ax3.set_box_aspect(1)
    # rotate labes axis
    ax3.xaxis.set_tick_params(rotation=45)
    # add legend
    ax3.legend(
        loc="upper left",
        title="Location",
    )

    # add base map
    if basemap:
        # check if contextily is installed
        if ctx is None:
            raise ImportError(
                "cannot import name contextily. To use this feature install pygef[map]."
            )

        # make sure the extent of the figure is not larger than 500 m by 500 m
        ax3.set_xbound(
            cpt_data.delivered_location.x - 250, cpt_data.delivered_location.x + 250
        )
        ax3.set_ybound(
            cpt_data.delivered_location.y - 250, cpt_data.delivered_location.y + 250
        )
        ctx.add_basemap(
            ax3,
            crs=cpt_data.delivered_location.srs_name,
            source=ctx.providers.OpenStreetMap.Mapnik,
        )

    # add watermark to plot
    plt.gcf().text(
        0,
        0,
        "created by pyGEF supported by CRUX Engineering Microservices BV (CEMS)",
        fontsize=10,
        color="gray",
        alpha=0.5,
    )

    return fig, axes



from __future__ import annotations

from typing import Any, Callable, TypeVar, cast

from lxml import etree

from pygef.bore import BoreData
from pygef.cpt import CPTData

T = TypeVar("T", CPTData, BoreData)

BaseParser = etree.XMLParser(resolve_entities=False, dtd_validation=False)


def read_xml(
    root: etree.Element,
    constructor: Callable[..., T],
    resolver_schema: dict[str, Any],
    payload_root: str,
) -> list[T]:
    namespaces = root.nsmap
    dd = root.find(payload_root, namespaces)

    # test xml has the correct payload root
    if dd is None:
        raise SyntaxError(
            "Could not read BRO XML file, incorrect payload root. Did you obtain the data from the BRO RESTful API?"
        )

    out: list[T] = []

    payloads = dd.findall("./*")
    for payload in payloads:
        # kwargs of attribute: value
        resolved = dict()

        for atrib, d in resolver_schema.items():
            d = cast(dict[str, Any], d)
            el = payload.find(d["xpath"], payload.nsmap)

            if el is not None:
                if "resolver" in d:
                    func = d["resolver"]

                    if "el-attr" in d:
                        el = getattr(el, d["el-attr"])

                    # ignore mypy error as it thinks we get a
                    # str from the dict
                    resolved[atrib] = func(el, namespaces=namespaces)
                else:
                    resolved[atrib] = el.text
            else:
                resolved[atrib] = None
        out.append(constructor(**resolved))

    return out



from __future__ import annotations

import copy
import pprint
from dataclasses import dataclass, field
from datetime import date
from typing import Any, List

import polars as pl

from pygef.common import Location, VerticalDatumClass, depth_to_offset


@dataclass(frozen=True)
class CPTData:
    """
    The CPT dataclass holds the information from the CPT object.

    Attributes:
        bro_id (str | None): BRO ID of the CPT.
        alias  (str | None): Alias of the CPT.
        research_report_date (date): research report date
        delivered_location (Location): delivered location in `EPSG:28992 - RD new`
        standardized_location (Location | None): standardized location in `EPSG:4326 - WGS 84`
        delivered_vertical_position_offset (float | None): delivered vertical position offset
        delivered_vertical_position_datum (str): research delivered vertical position datum
        delivered_vertical_position_reference_point (str): delivered vertical position reference point
        cpt_standard (str | None): cpt standard
        dissipationtest_performed (bool | None): dissipationtest performed
        quality_class (int | None): quality class. None if the value in the source file was invalid or missing.
        predrilled_depth (float): predrilled depth
        final_depth (float): final depth
        groundwater_level (float | None): groundwater level
        cpt_description (str): cpt description
        cpt_type (str): cpt type
        cone_surface_area (int): cone_surface_area
        cone_diameter (int | None): cone_diameter
        cone_surface_quotient (float | None): cone_surface_quotient
        cone_to_friction_sleeve_distance (int | None): cone_to_friction_sleeve_distance
        cone_to_friction_sleeve_surface_area (int | None): cone_to_friction_sleeve_surface_area
        cone_to_friction_sleeve_surface_quotient (float | None): cone_to_friction_sleeve_surface_quotient
        zlm_cone_resistance_before (float): zlm_cone_resistance_before
        zlm_cone_resistance_after (float): zlm_cone_resistance_after
        zlm_inclination_ew_before (int | None): zlm_inclination_ew_before
        zlm_inclination_ew_after (int | None): zlm_inclination_ew_after
        zlm_inclination_ns_before (int | None): zlm_inclination_ns_before
        zlm_inclination_ns_after (int | None): zlm_inclination_ns_after
        zlm_inclination_resultant_before (int | None): zlm_inclination_resultant_before
        zlm_inclination_resultant_after (int | None): zlm_inclination_resultant_after
        zlm_local_friction_before (float | None): zlm_local_friction_before
        zlm_local_friction_after (float | None): zlm_local_friction_after
        zlm_pore_pressure_u1_before (float | None): zlm_pore_pressure_u1_before
        zlm_pore_pressure_u2_before (float | None): zlm_pore_pressure_u2_before
        zlm_pore_pressure_u3_before (float | None): zlm_pore_pressure_u3_before
        zlm_pore_pressure_u1_after (float | None): zlm_pore_pressure_u1_after
        zlm_pore_pressure_u2_after (float | None): zlm_pore_pressure_u2_after
        zlm_pore_pressure_u3_after (float | None): zlm_pore_pressure_u3_after
        data (pl.DataFrame): DataFrame
            columns:

                - penetrationLength [m]
                - coneResistance [MPa]

            optional columns:

                - depthOffset [m wrt offset]
                    see delivered_vertical_position_datum for offset
                - depth [m]
                    penetrationLength corrected for inclination
                - correctedConeResistance [MPa]
                - netConeResistance [MPa]
                - coneResistanceRatio
                - localFriction [MPa]
                - frictionRatioComputed [%]
                - frictionRatio [%]
                - porePressureU1 [MPa]
                - porePressureU2 [MPa]
                - porePressureU3 [MPa]
                - inclinationResultant [degrees]
                - inclinationNS [degrees]
                - inclinationEW [degrees]
                - elapsedTime [seconds]
                - poreRatio [MPa]
                - soilDensity
                - porePressure
                - verticalPorePressureTotal
                - verticalPorePressureEffective
                - temperature [degrees celsius]
                - inclinationX [degrees]
                - inclinationY [degrees]
                - electricalConductivity [S/m]
                - magneticFieldStrengthX [nT]
                - magneticFieldStrengthY [nT]
                - magneticFieldStrengthZ [nT]
                - magneticFieldStrengthTotal [nT]
                - magneticInclination [degrees]
                - magneticDeclination [degrees]
    """

    # dispatch_document cpt
    bro_id: str | None
    research_report_date: date
    cpt_standard: str | None
    delivered_location: Location
    standardized_location: Location | None
    # conepenetrometersurvey
    dissipationtest_performed: bool | None
    quality_class: int | None
    predrilled_depth: float
    final_depth: float
    groundwater_level: float | None
    # conepenetrometer
    cpt_description: str
    cpt_type: str
    cone_surface_area: int
    cone_diameter: int | None
    cone_surface_quotient: float | None
    cone_to_friction_sleeve_distance: int | None
    cone_to_friction_sleeve_surface_area: int | None
    cone_to_friction_sleeve_surface_quotient: float | None
    # zero-load-measurement
    zlm_cone_resistance_before: float
    zlm_cone_resistance_after: float
    zlm_inclination_ew_before: int | None
    zlm_inclination_ew_after: int | None
    zlm_inclination_ns_before: int | None
    zlm_inclination_ns_after: int | None
    zlm_inclination_resultant_before: int | None
    zlm_inclination_resultant_after: int | None
    zlm_local_friction_before: float | None
    zlm_local_friction_after: float | None
    zlm_pore_pressure_u1_before: float | None
    zlm_pore_pressure_u2_before: float | None
    zlm_pore_pressure_u3_before: float | None
    zlm_pore_pressure_u1_after: float | None
    zlm_pore_pressure_u2_after: float | None
    zlm_pore_pressure_u3_after: float | None
    delivered_vertical_position_offset: float | None
    delivered_vertical_position_datum: VerticalDatumClass
    delivered_vertical_position_reference_point: str
    data: pl.DataFrame

    alias: str | None = field(default=None)

    def __post_init__(self):
        # post-processing of the data
        df = (
            self.data.lazy()
            .pipe(
                _calculate_depth_with_respect_to_offset,
                self.delivered_vertical_position_offset,
                self.data.columns,
            )
            .pipe(_calculate_friction_number, self.data.columns)
            .sort("penetrationLength", descending=False, nulls_last=False)
            .collect()
        )
        # bypass FrozenInstanceError
        object.__setattr__(self, "data", df)

    @property
    def columns(self) -> list[str]:
        """Columns names for the DataFrame"""
        return self.data.columns

    @property
    def groundwater_level_offset(self) -> float | None:
        """groundwater level wrt offset"""
        return depth_to_offset(
            self.groundwater_level, offset=self.delivered_vertical_position_offset
        )

    @property
    def predrilled_depth_offset(self) -> float | None:
        """predrilled depth wrt offset"""
        return depth_to_offset(
            self.predrilled_depth, offset=self.delivered_vertical_position_offset
        )

    @property
    def final_depth_offset(self) -> float | None:
        """final depth wrt offset"""
        return depth_to_offset(
            self.final_depth_offset, offset=self.delivered_vertical_position_offset
        )

    def __str__(self):
        return f"CPTData: {self.display_attributes()}"

    def attributes(self) -> dict[str, Any]:
        """
        Get the attributes
        """
        attribs = copy.copy(self.__dict__)
        attribs["data"] = attribs["data"].shape
        return attribs

    def display_attributes(self) -> str:
        """
        Get pretty formatted string representation of `CPTData.attributes``
        """
        return pprint.pformat(self.attributes())


def _calculate_friction_number(lf: pl.LazyFrame, columns: List[str]) -> pl.LazyFrame:
    """Post-process function for CPT data, creates a new column with the computed frictionRatio"""
    if "localFriction" in columns and "coneResistance" in columns:
        return lf.with_columns(
            (
                pl.col("localFriction")
                / pl.when(pl.col("coneResistance") == 0.0)
                .then(None)
                .otherwise(pl.col("coneResistance"))
                * 100.0
            ).alias("frictionRatioComputed")
        )
    return lf


def _calculate_depth_with_respect_to_offset(
    lf: pl.LazyFrame, offset: float | None, columns: List[str]
) -> pl.LazyFrame:
    """Post-process function for CPT data creates a new column with the elevation with respect to offset"""
    if offset is None:
        return lf
    yname = "depth" if "depth" in columns else "penetrationLength"
    return lf.with_columns((offset - pl.col(yname)).alias("depthOffset"))



from __future__ import annotations

import io
import os
from pathlib import Path
from typing import Any, Literal

from pygef.bore import BoreData
from pygef.broxml.parse_bore import read_bore as read_bore_xml
from pygef.broxml.parse_cpt import read_cpt as read_cpt_xml
from pygef.common import Location, VerticalDatumClass
from pygef.cpt import CPTData
from pygef.gef.parse_bore import _GefBore
from pygef.gef.parse_cpt import _GefCpt

GEF_ID = "#GEFID"


def is_gef_file(file: io.BytesIO | Path | str) -> bool:
    """
    gef files start with '#GEFID' so we check the content
    of the file
    """
    if isinstance(file, io.BytesIO):
        pos = file.tell()
        is_gef = file.read(6).decode().startswith(GEF_ID)
        file.seek(pos)
        return is_gef
    if os.path.exists(file):
        with open(file, errors="ignore") as f:
            return f.read(6).startswith(GEF_ID)
    if isinstance(file, str):
        return file[:6].startswith(GEF_ID)
    raise FileNotFoundError("Could not find the GEF file.")


def read_bore(
    file: io.BytesIO | Path | str,
    index: int = 0,
    engine: Literal["auto", "gef", "xml"] = "auto",
) -> BoreData:
    """
    Parse the bore file. Can either be BytesIO, Path or str

    :param file: bore file
    :param index: only valid for xml files
    :param engine: default is "auto". parsing engine.
        Please note that auto engine checks if the files starts with `#GEFID`.
    """
    if engine == "gef" or is_gef_file(file) and engine == "auto":
        if index > 0:
            raise ValueError("an index > 0 not supported for GEF files")
        if isinstance(file, io.BytesIO):
            return gef_bore_to_bore_data(_GefBore(string=file.read().decode()))
        if os.path.exists(file):
            return gef_bore_to_bore_data(_GefBore(path=file))
        else:
            return gef_bore_to_bore_data(_GefBore(string=file))
    return read_bore_xml(file)[index]


def read_cpt(
    file: io.BytesIO | Path | str,
    index: int = 0,
    engine: Literal["auto", "gef", "xml"] = "auto",
) -> CPTData:
    """
    Parse the cpt file. Can either be BytesIO, Path or str

    :param file: bore file
    :param index: only valid for xml files
    :param engine: default is "auto". parsing engine.
        Please note that auto engine checks if the files starts with `#GEFID`.
    """

    if engine == "gef" or is_gef_file(file) and engine == "auto":
        if index > 0:
            raise ValueError("an index > 0 not supported for GEF files")
        if isinstance(file, io.BytesIO):
            return gef_cpt_to_cpt_data(_GefCpt(string=file.read().decode()))
        if os.path.exists(file):
            return gef_cpt_to_cpt_data(_GefCpt(path=file))
        else:
            return gef_cpt_to_cpt_data(_GefCpt(string=file))
    return read_cpt_xml(file)[index]


def convert_height_system_to_vertical_datum(height_system: float) -> str:
    if height_system == 31000.0:
        return "nap"
    else:
        return str(int(height_system))


def gef_cpt_to_cpt_data(gef_cpt: _GefCpt) -> CPTData:
    kwargs: dict[str, Any] = {}

    kwargs["delivered_location"] = Location(
        # all gef files are RD new
        srs_name="urn:ogc:def:crs:EPSG::28992",
        x=gef_cpt.x,
        y=gef_cpt.y,
    )
    kwargs["standardized_location"] = None
    kwargs["bro_id"] = None
    kwargs["alias"] = gef_cpt.test_id
    kwargs["data"] = gef_cpt.df
    kwargs["research_report_date"] = gef_cpt.file_date
    kwargs["cpt_standard"] = None
    kwargs["groundwater_level"] = gef_cpt.groundwater_level
    kwargs["dissipationtest_performed"] = None
    kwargs["quality_class"] = gef_cpt.cpt_class
    kwargs["predrilled_depth"] = gef_cpt.pre_excavated_depth
    kwargs["final_depth"] = gef_cpt.end_depth_of_penetration_test
    kwargs["cpt_description"] = ""
    kwargs["cpt_type"] = gef_cpt.type_of_cone_penetration_test
    kwargs["cone_surface_area"] = gef_cpt.nom_surface_area_cone_tip
    kwargs["cone_diameter"] = None
    kwargs["cone_surface_quotient"] = gef_cpt.net_surface_area_quotient_of_the_cone_tip
    kwargs["cone_to_friction_sleeve_distance"] = (
        gef_cpt.distance_between_cone_and_centre_of_friction_casing
    )
    kwargs["cone_to_friction_sleeve_surface_area"] = None
    kwargs["cone_to_friction_sleeve_surface_quotient"] = (
        gef_cpt.net_surface_area_quotient_of_the_friction_casing
    )

    kwargs["zlm_cone_resistance_before"] = (
        gef_cpt.zero_measurement_cone_before_penetration_test
    )
    kwargs["zlm_cone_resistance_after"] = (
        gef_cpt.zero_measurement_cone_after_penetration_test
    )
    kwargs["zlm_inclination_ew_before"] = (
        gef_cpt.zero_measurement_inclination_ew_before_penetration_test
    )
    kwargs["zlm_inclination_ew_after"] = (
        gef_cpt.zero_measurement_inclination_ew_after_penetration_test
    )
    kwargs["zlm_inclination_ns_before"] = (
        gef_cpt.zero_measurement_inclination_ns_before_penetration_test
    )
    kwargs["zlm_inclination_ns_after"] = (
        gef_cpt.zero_measurement_inclination_ns_after_penetration_test
    )
    kwargs["zlm_inclination_resultant_before"] = None
    kwargs["zlm_inclination_resultant_after"] = None
    kwargs["zlm_local_friction_before"] = (
        gef_cpt.zero_measurement_friction_before_penetration_test
    )
    kwargs["zlm_local_friction_after"] = (
        gef_cpt.zero_measurement_friction_after_penetration_test
    )
    kwargs["zlm_pore_pressure_u1_before"] = (
        gef_cpt.zero_measurement_ppt_u1_before_penetration_test
    )
    kwargs["zlm_pore_pressure_u2_before"] = (
        gef_cpt.zero_measurement_ppt_u2_before_penetration_test
    )
    kwargs["zlm_pore_pressure_u3_before"] = (
        gef_cpt.zero_measurement_ppt_u3_before_penetration_test
    )
    kwargs["zlm_pore_pressure_u1_after"] = (
        gef_cpt.zero_measurement_ppt_u1_after_penetration_test
    )
    kwargs["zlm_pore_pressure_u2_after"] = (
        gef_cpt.zero_measurement_ppt_u2_after_penetration_test
    )
    kwargs["zlm_pore_pressure_u3_after"] = (
        gef_cpt.zero_measurement_ppt_u3_after_penetration_test
    )
    kwargs["delivered_vertical_position_offset"] = gef_cpt.zid
    kwargs["delivered_vertical_position_datum"] = VerticalDatumClass(
        str(int(gef_cpt.height_system))
    )

    # TODO! parse measurementtext 9 in gef?
    kwargs["delivered_vertical_position_reference_point"] = "unknown"

    return CPTData(**kwargs)


def gef_bore_to_bore_data(gef_bore: _GefBore) -> BoreData:
    kwargs: dict[str, Any] = {}

    kwargs["delivered_location"] = Location(
        # all gef files are RD new
        srs_name="urn:ogc:def:crs:EPSG::28992",
        x=gef_bore.x,
        y=gef_bore.y,
    )
    kwargs["standardized_location"] = None
    kwargs["bro_id"] = None
    kwargs["alias"] = gef_bore.test_id
    kwargs["groundwater_level"] = None
    kwargs["research_report_date"] = gef_bore.file_date
    kwargs["description_procedure"] = "unknown"
    kwargs["delivered_vertical_position_offset"] = gef_bore.zid
    kwargs["delivered_vertical_position_datum"] = "unknown"
    kwargs["delivered_vertical_position_reference_point"] = "unknown"
    kwargs["bore_rock_reached"] = None
    kwargs["final_bore_depth"] = None
    kwargs["final_sample_depth"] = None
    kwargs["bore_hole_completed"] = None
    kwargs["data"] = gef_bore.df
    return BoreData(**kwargs)
